"""
Some predefined network architectures for EEG decoding.
"""

from .deep4 import Deep4Net
from .eegnet import EEGNetv4, EEGNetv1
from .hybrid import HybridNet
from .shallow_fbcsp import ShallowFBCSPNet
from .eegresnet import EEGResNet
